@@ -691,8 +691,6 @@ class BaseModel(BaseClass):
                     *self.error_model.args, **self.error_model.kwargs
                 )
                 error_model.build(model, X_train, y_train)
-                if isinstance(error_model, BaseAggregateErrorModel):
-                    error_model.train_cv()
                 new_scores = error_model.calculate(X_test, y_pred_test)
                 if scores is None:
                     pred = y_pred_test
@@ -762,6 +760,8 @@ class BaseModel(BaseClass):
 
         import joblib
 
+        import joblib
+
         dataset_hash = joblib.hash(X) + joblib.hash(y)
 
         y_pred = self.predict(X)
@@ -1017,7 +1017,49 @@ class BaseSKLearnModel(BaseModel):
             self.regression_model._load(d["model"])
             self.model = self.regression_model
 
+class BaseReduction(BaseClass):
+    """
+    BaseReduction for applying dimensionality reduction on high-dimensional data
 
+    Parameters:
+        n_components (int): the number of components to keep
+
+    Methods:
+        fit: fit the model with input data
+        fit_transform: fit the model with and apply dimensionality reduction to input data
+        transform: apply dimensionality reduction to input data
+    """
+    def __init__(self):
+        pass
+
+    @abstractmethod
+    def fit(self, X):
+        pass
+    
+    @abstractmethod
+    def fit_transform(self, X):
+        pass
+
+    @abstractmethod
+    def transform(self, X):
+        pass
+    
+
+class BaseSKLearnReduction(BaseReduction):
+    """
+    Base class for creating sklearn dimensionality reduction
+
+    """
+    def fit(self, X):
+        return self.obj.fit(X)
+    
+    def fit_transform(self, X):
+        return self.obj.fit_transform(X)
+
+    def transform(self, X):
+        return self.obj.transform(X)
+
+    
 class BaseErrorModel(BaseClass):
     """Base class for error models.
 
@@ -1034,8 +1076,21 @@ class BaseErrorModel(BaseClass):
         score: returns confidence intervals on a dataset
     """
 
-    def __init__(self):
-        pass
+    @log_arguments
+    def __init__(self, ci: float = 0.8, 
+        method: str = "qbin",
+        window: int = 100,
+        bins: int = 10,
+        min_per_bin: int = 5,
+        log=True, 
+        **kwargs):
+        
+        self.ci = ci
+        self.method = method
+        self.window = window
+        self.bins = bins
+        self.min_per_bin = min_per_bin
+        
 
     def build(
         self,
@@ -1076,6 +1131,7 @@ class BaseErrorModel(BaseClass):
         self,
         X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
         y: Union[np.ndarray, list, pd.Series],
+        ci: float = None,
         **kwargs,
     ):
         """Fits confidence scores to an external dataset
@@ -1084,42 +1140,50 @@ class BaseErrorModel(BaseClass):
             X (array-like): features, smiles
             y (array-like): true values
         """
+        if ci is None:
+            ci = self.ci
+            
+        X = oce.SMILESRepresentation().convert(X)
         y_pred = np.array(self.model.predict(X)).flatten()
         residuals = np.abs(np.array(y) - y_pred)
-
         scores = np.array(self.calculate(X, y_pred))
-        self._fit(residuals, scores, **kwargs)
 
-    def fit_cv(
-        self,
-        n_splits: int = 10,
-        test_size_range: Union[tuple, list] = (0.05, 0.5),
-        **kwargs,
-    ):
+        self._fit(residuals, scores, quantile = ci, **kwargs)
+
+    def fit_cv(self,  X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
+        y: Union[np.ndarray, list, pd.Series], n_splits: int = 5, ci: float = None, **kwargs):
         """Fits confidence scores to the training dataset via cross validation.
 
         Args:
-            n_splits (int): number of cross validations
-            test_size_range (tuple or list): range of test split proportions
+            n_splits (int): Number of cross validation splits, default 5
         """
+        
+        if ci is None:
+            ci = self.ci
+            
+        from sklearn.model_selection import KFold
+        from sklearn.calibration import calibration_curve
+        
+        X = np.array(oce.SMILESRepresentation().convert(X))
+        if issubclass(type(y), pd.Series):
+            y = y.values
+        
         residuals = None
         scores = None
-        splits = np.linspace(*test_size_range, n_splits)
+        kf = KFold(n_splits=n_splits)
 
-        from sklearn.model_selection import train_test_split
-
-        for test_size in splits:
-            print("cross validating with split={}".format(np.round(test_size, 3)))
-            X_train, X_test, y_train, y_test = train_test_split(
-                self.X_train, self.y_train, test_size=test_size, random_state=0
-            )
+        for train_index, test_index in kf.split(X):
+            X_train, X_test = X[train_index], X[test_index]
+            y_train, y_test = y[train_index], y[test_index]
             model = self.model.copy()
             model.fit(X_train, y_train)
+            
             y_pred_test = np.array(model.predict(X_test)).flatten()
+            pred_error = np.abs(y_test - y_pred_test)
             if residuals is None:
-                residuals = np.abs(y_test - y_pred_test)
+                residuals = pred_error
             else:
-                residuals = np.concatenate((residuals, np.abs(y_test - y_pred_test)))
+                residuals = np.concatenate((residuals, pred_error))
 
             em = type(self)(*self.args, **self.kwargs)
             em.build(model, X_train, y_train)
@@ -1129,17 +1193,13 @@ class BaseErrorModel(BaseClass):
             else:
                 scores = np.concatenate((scores, new_scores))
 
-        self._fit(residuals, scores, **kwargs)
+        self._fit(residuals, scores, quantile = ci, **kwargs)
 
     def _fit(
         self,
         residuals: np.ndarray,
         scores: np.ndarray,
-        quantile: float = 0.8,
-        method: str = "roll",
-        window: int = 100,
-        bins: int = 10,
-        min_per_bin: int = 5,
+        quantile: float = 0.95,
         filename: str = "figure.png",
     ):
         """Fits confidence scores to residuals.
@@ -1154,46 +1214,40 @@ class BaseErrorModel(BaseClass):
             min_per_bin (int): minimum number of instances per bin
             filename (str): save destination of the fitted plot
         """
-        if method == "bin":
-            bin_labels = pd.cut(scores, bins, labels=False)
+        if self.method == "bin":
+            bin_labels = pd.cut(scores, self.bins, labels=False)
             X = []
             y = []
-            for i in range(bins):
+            for i in range(self.bins):
                 ith_bin = bin_labels == i
-                if np.sum(ith_bin) >= min_per_bin:
+                if np.sum(ith_bin) >= self.min_per_bin:
                     X.append(np.mean(scores[ith_bin]))
                     y.append(pd.Series(residuals[ith_bin]).quantile(quantile))
             X = np.array(X)
             y = np.array(y)
-        elif method == "qbin":
-            bin_labels = pd.qcut(scores, bins, labels=False, duplicates="drop")
-            X = np.array(
-                [np.mean(scores[bin_labels == i]) for i in range(max(bin_labels) + 1)]
-            )
-            y = np.array(
-                [
-                    pd.Series(residuals[bin_labels == i]).quantile(quantile)
-                    for i in range(max(bin_labels) + 1)
-                ]
-            )
-        elif method == "roll":
+        elif self.method == "qbin":
+            bin_labels = pd.qcut(scores, self.bins, labels=False, duplicates="drop")
+            n_labels = int(max(bin_labels) + 1)
+            X = np.array([np.mean(scores[bin_labels == i]) for i in range(n_labels)])
+            y = np.array([pd.Series(residuals[bin_labels == i]).quantile(quantile) for i in range(n_labels)])
+        elif sself.method == "roll":
             results = list(zip(scores, residuals))
             scores, residuals = zip(*sorted(results))
-            X = pd.Series(scores).rolling(window).mean()
-            y = pd.Series(residuals).rolling(window).quantile(quantile)
+            X = pd.Series(scores).rolling(self.window).mean()
+            y = pd.Series(residuals).rolling(self.window).quantile(quantile)
             X = np.array(X[~np.isnan(X)])
             y = np.array(y[~np.isnan(y)])
         else:
-            raise NameError("method {} is not recognized".format(method))
+            raise NameError("method {} is not recognized".format(self.method))
 
         import matplotlib.pyplot as plt
         from scipy.optimize import curve_fit
 
         funcs = [
-            lambda x, a, b: a * np.exp(b * x),
-            lambda x, a, b: a * np.log(x) + b,
-            lambda x, a, b: a * x**b,
-            lambda x, a, b: a * x + b,
+            lambda x, a, b, c: a * np.exp(b * x) + c,
+            lambda x, a, b, c: a * np.log(x + b) + c,
+            lambda x, a, b, c: a * x ** b + c,
+            lambda x, a, b: a * x + b
         ]
         min_mse = None
         for func in funcs:
@@ -1210,15 +1264,12 @@ class BaseErrorModel(BaseClass):
                 opt_func = func
                 opt_popt = popt
                 min_mse = mse
-        X_min, X_max = np.min(X), np.max(X)
-        self.reg = np.vectorize(
-            lambda x: opt_func(max(min(x, X_max), X_min), *opt_popt)
-        )
+        floor, ceil = np.min(residuals), np.max(residuals)
+        self.reg = np.vectorize(lambda x: max(min(opt_func(x, *opt_popt), ceil), floor))
 
+        sorted_scores = np.sort(scores)
         plt.xlabel(self.__class__.__name__)
         plt.ylabel("Absolute Error")
-        sorted_scores = np.sort(scores)
-
         plt.scatter(scores, residuals, c="black", alpha=0.2)
         plt.scatter(X, y, c="blue")
         plt.plot(sorted_scores, self.reg(sorted_scores), c="red")
@@ -1226,11 +1277,7 @@ class BaseErrorModel(BaseClass):
 
         self.residuals = residuals
         self.scores = scores
-        self.method = method
-        self.bins = bins
-        self.window = window
         self.quantile = quantile
-        self.min_per_bin = min_per_bin
         self.filename = filename
 
     def score(
@@ -1304,162 +1351,6 @@ class BaseErrorModel(BaseClass):
             self._fit(
                 residuals,
                 scores,
-                method=method,
-                bins=bins,
-                window=window,
                 quantile=quantile,
-                min_per_bin=min_per_bin,
                 filename=filename,
-            )
-
-
-class BaseAggregateErrorModel(BaseErrorModel):
-    """Base class for aggregate error models.
-
-    Parameters:
-        error_models (BaseErrorModel or list of BaseErrorModel): list of error models to be aggregated
-    """
-
-    @log_arguments
-    def __init__(
-        self, error_models: Union[BaseErrorModel, List[BaseErrorModel]], **kwargs
-    ):
-        if not isinstance(error_models, list):
-            error_models = [error_models]
-        self.error_models = error_models
-        self.kwargs = kwargs
-
-    def build(
-        self,
-        model: BaseModel,
-        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
-        y: Union[np.ndarray, list, pd.Series],
-    ):
-        for error_model in self.error_models:
-            error_model.build(model, X, y)
-        super().build(model, X, y)
-
-    def train(
-        self,
-        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
-        y: Union[np.ndarray, list, pd.Series],
-    ):
-        """Trains the aggregate model to a validation dataset.
-
-        Args:
-            X (array-like): features, smiles
-            y (array-like): true values
-        """
-        y_pred = np.array(self.model.predict(X)).flatten()
-        residuals = np.abs(np.array(y) - y_pred)
-
-        scores = np.transpose(
-            np.stack([self.get_scores(em, X, y_pred) for em in self.error_models])
-        )
-        self._train(scores, residuals)
-
-    def train_cv(
-        self, n_splits: int = 9, test_size_range: Union[tuple, list] = (0.1, 0.9)
-    ):
-        """Trains the aggregate model to the training dataset via cross validation.
-
-        Args:
-            n_splits (int): number of cross validations
-            test_size_range (tuple or list): range of test split proportions
-        """
-        residuals = None
-        scores = None
-        splits = np.linspace(*test_size_range, n_splits)
-
-        from sklearn.model_selection import train_test_split
-
-        for test_size in splits:
-            print("cross validating with split={}".format(np.round(test_size, 3)))
-            X_train, X_test, y_train, y_test = train_test_split(
-                self.X_train, self.y_train, test_size=test_size, random_state=0
-            )
-            model = self.model.copy()
-            model.fit(X_train, y_train)
-            y_pred_test = np.array(model.predict(X_test)).flatten()
-            if residuals is None:
-                residuals = np.abs(y_test - y_pred_test)
-            else:
-                residuals = np.concatenate((residuals, np.abs(y_test - y_pred_test)))
-
-            new_scores = []
-            for error_model in self.error_models:
-                em = type(error_model)(*error_model.args, **error_model.kwargs)
-                em.build(model, X_train, y_train)
-                new_scores.append(self.get_scores(em, X_test, y_pred_test))
-            new_scores = np.transpose(np.stack(new_scores))
-            if scores is None:
-                scores = new_scores
-            else:
-                scores = np.concatenate((scores, new_scores))
-
-        self._train(scores, residuals)
-
-    def calculate(
-        self,
-        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
-        y_pred: np.ndarray,
-        **kwargs,
-    ) -> np.ndarray:
-        """Computes aggregate error model score from inputs.
-
-        Args:
-            X: features, smiles
-            y_pred: predicted values
-        """
-        assert hasattr(self, "aggregate_model"), "aggregate model not yet trained"
-
-        scores = [self.get_scores(em, X, y_pred) for em in self.error_models]
-        scores = np.transpose(np.stack(scores))
-        return self._predict(scores, **kwargs)
-
-    def get_scores(
-        self,
-        em: BaseErrorModel,
-        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
-        y_pred: np.ndarray,
-    ) -> np.ndarray:
-        """Gets confidence scores from a model, inputs, and keyword arguments.
-
-        Args:
-            em (BaseErrorModel): error model
-            X: features, smiles
-            y_pred: predicted values
-        """
-        return em.calculate(X, y_pred)
-
-    @abstractmethod
-    def _train(self, scores: np.ndarray, residuals: np.ndarray, **kwargs):
-        """To be implemented by the child class; trains the aggregate model based on residuals
-        and confidence scores and assigns the trained model to `self.aggregate_model`.
-
-        Args:
-            scores (np.ndarray): confidence scores
-            residuals (np.ndarray): residuals
-        """
-        pass
-
-    @abstractmethod
-    def _predict(self, scores: np.ndarray, **kwargs) -> np.ndarray:
-        """To be implemented by the child class; predicts from the aggregate model based on
-        confdience scores.
-
-        Args:
-            scores (np.ndarray): confidence scores
-        """
-        pass
-
-    def _save(self) -> dict:
-        d = super()._save()
-        if hasattr(self, "aggregate_model"):
-            d.update({"aggregate_model": saves(self.aggregate_model)})
-        return d
-
-    def _load(self, d) -> None:
-        super()._load(d)
-        if "aggregate_model" in d.keys():
-            self.aggregate_model = loads(d["aggregate_model"])
+            )
\ No newline at end of file
