@@ -166,6 +166,7 @@ class RandomForestModel(BaseSKLearnModel):
         class_weight=None,
         bootstrap=True,
         n_estimators=100,
+        random_state=None,
         **kwargs
     ):
         if not class_weight is None:
@@ -180,8 +181,9 @@ class RandomForestModel(BaseSKLearnModel):
             max_features=max_features,
             min_samples_split=4,
             min_samples_leaf=4,
-            max_depth=max_depth,
+            max_depth = max_depth,
             n_jobs=-1,
+            random_state=random_state
         )
         classifier = RandomForestClassifier(
             max_features=max_features,
@@ -189,6 +191,7 @@ class RandomForestModel(BaseSKLearnModel):
             min_samples_leaf=4,
             n_estimators=n_estimators,
             n_jobs=-1,
+            random_state=random_state,
             criterion=criterion,
             bootstrap=bootstrap,
             max_depth=max_depth,
@@ -553,9 +556,10 @@ class XGBoostModel(BaseSKLearnModel, BaseObject):
         **kwargs
     ):
         from xgboost import XGBClassifier, XGBRegressor
+        import torch
 
         regressor = XGBRegressor(
-            tree_method="gpu_hist",
+            tree_method="gpu_hist" if torch.cuda.is_available() else "hist",
             n_estimators=n_estimators,
             max_depth=max_depth,
             max_leaves=max_leaves,
@@ -565,7 +569,7 @@ class XGBoostModel(BaseSKLearnModel, BaseObject):
         )
 
         classifier = XGBClassifier(
-            tree_method="gpu_hist",
+            tree_method="gpu_hist" if torch.cuda.is_available() else "hist",
             n_estimators=n_estimators,
             max_depth=max_depth,
             max_leaves=max_leaves,
@@ -602,6 +606,7 @@ class ZWK_XGBoostModel(BaseSKLearnModel, BaseObject):
         self, representation, n_iter=100, scoring=None, verbose=2, cv=5, **kwargs
     ):
         from xgboost import XGBClassifier, XGBRegressor
+        import torch
 
         if scoring == "spearman":
             from scipy.stats import spearmanr
@@ -615,10 +620,10 @@ class ZWK_XGBoostModel(BaseSKLearnModel, BaseObject):
             scoring = spearman
 
         regressor = self.autofit(
-            XGBRegressor(tree_method="gpu_hist"), n_iter, cv, scoring, verbose
+            XGBRegressor(tree_method="gpu_hist" if torch.cuda.is_available() else "hist"), n_iter, cv, scoring, verbose
         )
         classifier = self.autofit(
-            XGBClassifier(tree_method="gpu_hist"), n_iter, cv, scoring, verbose
+            XGBClassifier(tree_method="gpu_hist" if torch.cuda.is_available() else "hist"), n_iter, cv, scoring, verbose
         )
 
         super().__init__(representation, regressor, classifier, log=False, **kwargs)
