@@ -1,14 +1,10 @@
-from cmath import exp
-from tkinter import N
+""" Techniques for quantifying uncertainty and estimating confidence intervals for all oce models.
+"""
 
 import numpy as np
 import pandas as pd
 from rdkit.Chem import AllChem
 from rdkit.DataStructs.cDataStructs import BulkTanimotoSimilarity
-from sklearn.cross_decomposition import PLSRegression
-from sklearn.decomposition import PCA
-from sklearn.neighbors import NearestNeighbors
-from sklearn.preprocessing import OneHotEncoder
 
 import olorenchemengine as oce
 
@@ -17,6 +13,120 @@ from .basics import *
 from .dataset import *
 from .representations import *
 
+class BaseEnsembleModel(BaseErrorModel):
+    """ BaseEnsembleModel is the base class for error models that estimate
+        uncertainty based on the variance of an ensemble of models.
+    """
+
+    @log_arguments
+    def __init__(self, ensemble_model = None, n_ensembles = 16, log=True, **kwargs):      
+        self.ensemble_model = ensemble_model
+        self.n_ensembles = n_ensembles
+        super().__init__(log=False, **kwargs)
+
+    def build(
+        self,
+        model: BaseModel,
+        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
+        y: Union[np.ndarray, list, pd.Series],
+    ):
+        super().build(model, X, y)
+        if self.ensemble_model is None:
+            self.ensemble_model = self.model.copy()
+        self.ensembles = []
+    
+    def calculate(self, X, y_pred):
+        predictions = np.stack([model.predict(X) for model in tqdm(self.ensembles)])
+        return np.var(predictions, axis=0)
+
+    def _save(self) -> dict:
+        d = super()._save()
+        if hasattr(self, "ensembles"):
+            d.update({"ensembles": [model._save() for model in self.ensembles]})
+        return d
+
+    def _load(self, d) -> None:
+        super()._load(d)
+        if "ensembles" in d.keys():
+            self.ensembles = [model._load() for model in d["ensembles"]]
+    
+class BootstrapEnsemble(BaseEnsembleModel):
+    """ BootstrapEnsemble estimates uncertainty based on the variance of several
+        models trained on bootstrapped samples of the training data.
+
+        Parameters:
+            ensemble_model (BaseModel): Model used for ensembling. Defaults to the same as the original model.
+            n_ensembles (int): Number of ensembles
+            bootstrap_size (float): Proportion of training data to train each ensemble model
+    
+    Example
+    ------------------------------
+    import olorenchemengine as oce
+
+    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
+    model.fit_cv(train["Drug"], train["Y"], error_model = oce.BootstrapEnsemble(n_ensembles = 10))
+    model.predict(test["Drug"], return_ci = True)
+    ------------------------------
+    """
+    
+    @log_arguments
+    def __init__(self, ensemble_model = None, n_ensembles = 16, bootstrap_size = 0.25, log=True, **kwargs):
+        self.bootstrap_size = bootstrap_size
+        super().__init__(ensemble_model = ensemble_model, n_ensembles = n_ensembles,log=False, **kwargs)
+
+    def build(
+        self,
+        model: BaseModel,
+        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
+        y: Union[np.ndarray, list, pd.Series],
+    ):
+        super().build(model, X, y)
+
+        from sklearn.model_selection import train_test_split
+        
+        for _ in tqdm(range(self.n_ensembles)):
+            X_train, X_test, y_train, y_test = train_test_split(
+                self.X_train, self.y_train, train_size=self.bootstrap_size
+            )
+            ensemble_model = self.ensemble_model.copy()
+            ensemble_model.fit(X_train, y_train)
+            self.ensembles.append(ensemble_model)
+
+class RandomForestEnsemble(BaseEnsembleModel):
+    """ RandomForestEnsemble estimates uncertainty based on the variance of several
+        random forest models initialized to different random states.
+
+        Parameters:
+            ensemble_model (BaseModel): Model used for ensembling. Defaults to the same as the original model.
+            n_ensembles (int): Number of ensembles
+    
+    Example
+    ------------------------------
+    import olorenchemengine as oce
+
+    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
+    model.fit_cv(train["Drug"], train["Y"], error_model = oce.RandomForestEnsemble(n_ensembles = 10))
+    model.predict(test["Drug"], return_ci = True)
+    ------------------------------
+    """
+    
+    def build(
+        self,
+        model: BaseModel,
+        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
+        y: Union[np.ndarray, list, pd.Series],
+        **kwargs
+    ):
+        super().build(model, X, y)
+        
+        for n in tqdm(range(self.n_ensembles)):
+            ensemble_model = RandomForestModel(
+                oce.MorganVecRepresentation(radius=2, nbits=2048), 
+                random_state=n, 
+                **kwargs
+            )
+            ensemble_model.fit(self.X_train, self.y_train)
+            self.ensembles.append(ensemble_model)
 
 class BaseFingerprintModel(BaseErrorModel):
     """BaseFingerprintModel is the base class for error models that require the
@@ -68,17 +178,19 @@ class SDC(BaseFingerprintModel):
     """
 
     @log_arguments
-    def __init__(self, a=3):
+    def __init__(self, a=3, log=True, **kwargs):
         self.a = a
+        super().__init__(log=False, **kwargs)
 
     def calculate(self, X, y_pred):
+        X = SMILESRepresentation().convert(X)
         def sdc(smi):
             mol = Chem.MolFromSmiles(smi)
             ref_fp = AllChem.GetMorganFingerprint(mol, 2)
             TD = 1 - np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
             return np.sum(np.exp(-self.a * TD / (1 - TD)))
 
-        return np.array([sdc(smi) for _, smi in tqdm(np.ndenumerate(X))])
+        return np.array([sdc(smi) for smi in tqdm(X)])
 
 
 class TargetDistDC(SDC):
@@ -100,15 +212,19 @@ class TargetDistDC(SDC):
     """
 
     def calculate(self, X, y_pred):
+        X = SMILESRepresentation().convert(X)
         def dist(smi, pred):
             mol = Chem.MolFromSmiles(smi)
             ref_fp = AllChem.GetMorganFingerprint(mol, 2)
             TD = 1 - np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
             DC = np.exp(-self.a * TD / (1 - TD))
-            return np.sqrt(np.dot(DC, np.abs(self.y_train - pred)) / np.sum(DC))
+            error = np.abs(self.y_train - pred)
+            if np.sum(DC) == 0:
+                return np.sqrt(np.mean(error))
+            return np.sqrt(np.dot(DC, error) / np.sum(DC))
 
         X = np.array(X).flatten()
-        return np.array([dist(smi, y_pred[i]) for i, smi in tqdm(np.ndenumerate(X))])
+        return np.array([dist(smi, pred) for smi, pred in tqdm(zip(X, y_pred))])
 
 
 class TrainDistDC(SDC):
@@ -130,6 +246,7 @@ class TrainDistDC(SDC):
     """
 
     def calculate(self, X, y_pred):
+        X = SMILESRepresentation().convert(X)
         residuals = np.abs(self.y_train - self.y_pred_train)
 
         def dist(smi):
@@ -137,9 +254,11 @@ class TrainDistDC(SDC):
             ref_fp = AllChem.GetMorganFingerprint(mol, 2)
             TD = 1 - np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
             DC = np.exp(-self.a * TD / (1 - TD))
+            if np.sum(DC) == 0:
+                return np.sqrt(np.mean(residuals))
             return np.sqrt(np.dot(DC, residuals) / np.sum(DC))
 
-        return np.array([dist(smi) for _, smi in tqdm(np.ndenumerate(X))])
+        return np.array([dist(smi) for smi in tqdm(X)])
 
 
 class KNNSimilarity(BaseFingerprintModel):
@@ -161,17 +280,19 @@ class KNNSimilarity(BaseFingerprintModel):
     """
 
     @log_arguments
-    def __init__(self, k=5):
+    def __init__(self, k=5, log=True, **kwargs):
         self.k = k
+        super().__init__(log=False, **kwargs)
 
     def calculate(self, X, y_pred):
+        X = SMILESRepresentation().convert(X)
         def mean_sim(smi):
             mol = Chem.MolFromSmiles(smi)
             ref_fp = AllChem.GetMorganFingerprint(mol, 2)
             similarity = np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
             return np.mean(sorted(similarity)[-self.k :])
 
-        return np.array([mean_sim(smi) for _, smi in tqdm(np.ndenumerate(X))])
+        return np.array([mean_sim(smi) for smi in tqdm(X)])
 
 
 class TargetDistKNN(KNNSimilarity):
@@ -194,18 +315,19 @@ class TargetDistKNN(KNNSimilarity):
     """
 
     def calculate(self, X, y_pred):
+        X = SMILESRepresentation().convert(X)
         def dist(smi, pred):
             mol = Chem.MolFromSmiles(smi)
             ref_fp = AllChem.GetMorganFingerprint(mol, 2)
             similarity = np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
-            mae = np.abs(self.y_train - pred)
-            idxs = np.argsort(similarity)[-self.k :]
-            return np.sqrt(
-                np.dot(similarity[idxs], mae[idxs]) / np.sum(similarity[idxs])
-            )
+            error = np.abs(self.y_train - pred)
+            idxs = np.argsort(similarity)[-self.k:]
+            if np.sum(similarity[idxs]) == 0:
+                return np.sqrt(np.mean(error[idxs]))
+            return np.sqrt(np.dot(similarity[idxs], error[idxs]) / np.sum(similarity[idxs]))
 
         X = np.array(X).flatten()
-        return np.array([dist(smi, y_pred[i]) for i, smi in tqdm(np.ndenumerate(X))])
+        return np.array([dist(smi, pred) for smi, pred in tqdm(zip(X, y_pred))])
 
 
 class TrainDistKNN(KNNSimilarity):
@@ -228,18 +350,19 @@ class TrainDistKNN(KNNSimilarity):
     """
 
     def calculate(self, X, y_pred):
+        X = SMILESRepresentation().convert(X)
         residuals = np.abs(self.y_train - self.y_pred_train)
 
         def dist(smi):
             mol = Chem.MolFromSmiles(smi)
             ref_fp = AllChem.GetMorganFingerprint(mol, 2)
             similarity = np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
-            idxs = np.argsort(similarity)[-self.k :]
-            return np.sqrt(
-                np.dot(similarity[idxs], residuals[idxs]) / np.sum(similarity[idxs])
-            )
+            idxs = np.argsort(similarity)[-self.k:]
+            if np.sum(similarity[idxs]) == 0:
+                return np.sqrt(np.mean(residuals[idxs]))
+            return np.sqrt(np.dot(similarity[idxs], residuals[idxs]) / np.sum(similarity[idxs]))
 
-        return np.array([dist(smi) for _, smi in tqdm(np.ndenumerate(X))])
+        return np.array([dist(smi) for smi in tqdm(X)])
 
 
 class Predicted(BaseErrorModel):
@@ -256,11 +379,6 @@ class Predicted(BaseErrorModel):
     model.predict(test["Drug"], return_ci = True)
     ------------------------------
     """
-
-    @log_arguments
-    def __init__(self):
-        pass
-
     def calculate(self, X, y_pred):
         return y_pred
 
@@ -285,8 +403,9 @@ class ADAN(BaseErrorModel):
     """
 
     @log_arguments
-    def __init__(self, criterion: str):
+    def __init__(self, criterion: str, log=True, **kwargs):
         self.criterion = criterion
+        super().__init__(log=False, **kwargs)
 
     def build(
         self,
@@ -306,11 +425,15 @@ class ADAN(BaseErrorModel):
         max_components = min(100, min(self.X_train.shape))
 
         if dim_reduction == "pls":
+            from sklearn.cross_decomposition import PLSRegression
+
             self.reduction = PLSRegression(n_components=max_components)
             self.reduction.fit(self.X_train, self.y_train)
             x_var = np.var(self.reduction.x_scores_, axis=0)
             x_var /= np.sum(x_var)
         elif dim_reduction == "pca":
+            from sklearn.decomposition import PCA
+
             self.reduction = PCA(n_components=max_components)
             self.reduction.fit(self.X_train, self.y_train)
             x_var = self.reduction.explained_variance_ratio_
@@ -328,6 +451,8 @@ class ADAN(BaseErrorModel):
         self.Xp_mean = np.mean(self.Xp_train, axis=0)
         self.y_mean = np.mean(self.y_train)
 
+        from sklearn.neighbors import NearestNeighbors
+
         nbrs = NearestNeighbors(n_neighbors=2).fit(self.Xp_train)
         distances, indices = nbrs.kneighbors(self.Xp_train)
 
@@ -357,6 +482,7 @@ class ADAN(BaseErrorModel):
         }
 
     def calculate_full(self, X):
+        X = SMILESRepresentation().convert(X)
         criteria = ["A", "B", "C", "D", "E", "F"]
         y_pred = np.array(self.model.predict(X)).flatten()
         X = self.preprocess(X)
@@ -373,6 +499,7 @@ class ADAN(BaseErrorModel):
         self.results = pd.DataFrame(self.results)
 
     def calculate(self, X, y_pred):
+        X = SMILESRepresentation().convert(X)
         """Calcualtes confidence scores."""
         X = self.preprocess(X)
         Xp = self.reduction.transform(X)[:, : self.n_components]
@@ -380,6 +507,7 @@ class ADAN(BaseErrorModel):
         return self._calculate(X, Xp, y_pred, self.criterion)
 
     def _calculate(self, X, Xp, y_pred, criterion: str, standardize: bool = True):
+        from sklearn.neighbors import NearestNeighbors
         if criterion in ("A", "A_raw"):
             dist = np.linalg.norm(Xp - self.Xp_mean, axis=1)
         elif criterion in ("B", "B_raw"):
@@ -463,6 +591,7 @@ class ADAN(BaseErrorModel):
             neighbor_thresh (float): fraction of closest training queries to consider
         """
         n_neighbors = int(self.X_train.shape[0] * neighbor_thresh) + n_drop
+        from sklearn.neighbors import NearestNeighbors
         nbrs = NearestNeighbors(n_neighbors=n_neighbors).fit(self.Xp_train)
         distances, indices = nbrs.kneighbors(Xp)
 
@@ -471,180 +600,73 @@ class ADAN(BaseErrorModel):
 
         return np.sqrt(y_mse).flatten()
 
-
-class RandomForestErrorModel(BaseAggregateErrorModel):
-    """RandomForestErrorModel is an aggregate error model that predicts error
-        bars based on an aggregate score determined by a random forest model
-        trained on several BaseErrorModels.
+class AggregateErrorModel(BaseErrorModel):
+    """ AggregateErrorModel estimates uncertainty by aggregating ucertainty scores from
+        several different BaseErrorModels.
 
         Parameters:
-            error_models (Union[BaseErrorModel, List[BaseErrorModel]]): a list
-                of error models whose scores will be features for the random
-                forest model
-
+            error_models (list of BaseErrorModel): list of error models to be aggregated
+            reduction (BaseReduction): reduction method used to aggregate uncertainty scores
+    
     Example
     ------------------------------
     import olorenchemengine as oce
 
     model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
-    model.fit_cv(train["Drug"], train["Y"], error_model = oce.RandomForestErrorModel([oce.SDC(), oce.KNNSimilarity()])
-    model.predict(test["Drug"], return_ci = True)
+    model.fit(train["Drug"], train["Y"])
+    error_model = oce.AggregateErrorModel(error_models = [oce.TargetDistDC(), oce.TrainDistDC()], reduction = oce.FactorAnalysis())
+    error_model.build(model, train["Drug"], train["Y"])
+    error_model.fit(valid["Drug"], valid["Y"])
+    error_model.score(test["Drug"])
     ------------------------------
     """
 
-    def _train(self, scores, residuals):
-        self.aggregate_model = RandomForestRegressor(**self.kwargs)
-        self.aggregate_model.fit(scores, residuals)
-
-    def _predict(self, scores, **kwargs):
-        return self.aggregate_model.predict(scores, **kwargs)
-
-
-class LinearRegressionErrorModel(BaseAggregateErrorModel):
-    """LinearRegressionErrorModel is an aggregate error model that predicts error
-        bars based on a linear combination score from several BaseErrorModels.
-
-        Parameters:
-            error_models (Union[BaseErrorModel, List[BaseErrorModel]]): a list
-                of error models whose scores will be features for the random
-                forest model
-
-    Example
-    ------------------------------
-    import olorenchemengine as oce
-
-    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
-    model.fit_cv(train["Drug"], train["Y"], error_model = oce.LinearRegressionErrorModel([oce.SDC(), oce.KNNSimilarity()])
-    model.predict(test["Drug"], return_ci = True)
-    ------------------------------
-    """
-
-    def _train(self, scores, residuals):
-        self.aggregate_model = LinearRegression(**self.kwargs)
-        self.aggregate_model.fit(scores, residuals)
-
-    def _predict(self, scores, **kwargs):
-        return self.aggregate_model.predict(scores, **kwargs)
-
-
-class BaseDomainApplicability(BaseClass):
-    """Depricated. See BaseErrorModel.
-
-    Base class for analyzing domain applicability of models and estimating
-    prediction uncertainty.
-
-    Parameters:
-        model (BaseModel): trained model
-        rep (BaseCompoundVecRepresentation): preprocessing representation
-
-    Methods:
-        _fit: fits a method to a trained model and preprocessed training dataset, to be implemented internally by child classes
-        fit: fits a method to a trained model and training dataset
-        _test evaluates prediction uncertainty on a preprocessed testing dataset, to be implemented internally by child classes
-        test: evaluates prediction uncertainty on a testing dataset
-    """
-
     @log_arguments
-    def __init__(self, model: BaseModel, rep: BaseCompoundVecRepresentation = None):
-        self.model = model
-        self.rep = rep
-        self.results = None
+    def __init__(self, error_models: List[BaseErrorModel], reduction: BaseReduction, log=True,
+                 **kwargs):
+        if not isinstance(error_models, list):
+            raise TypeError("error_models must be a list")
+        self.error_models = error_models
+        self.reduction = reduction
+        super().__init__(log=False, **kwargs)
 
-    @abstractmethod
-    def _fit(self, X, y, y_pred, **kwargs):
-        """To be implemented by the child class; fits the method on the internally stored training dataset.
-
-        Note: _fit shouldn"t be directly called by the user, rather it should be called indirectly via the fit method.
-
-        Parameters:
-            X (np.ndarray): features, SMILES
-            y (np.ndarray): values
-            y_pred (np.ndarray): predicted values
-        """
-        pass
-
-    def fit(self, dataset: BaseDataset, use_entire_dataset: bool = False, **kwargs):
-        """Fits model uncertainty to a training dataset. Calls the _fit method on the provided dataset.
-
-        Parameters:
-            dataset (BaseDataset): dataset used for fitting the method
-            use_entire_dataset (bool): whether to use the entire dataset or only the training data
-        """
-        if use_entire_dataset:
-            X = dataset.entire_dataset[0]
-            y = np.array(dataset.entire_dataset[1])
-        else:
-            X = dataset.train_dataset[0]
-            y = np.array(dataset.train_dataset[1])
-
-        y_pred = np.array(self.model.predict(X))
-
-        self._fit(X, y, y_pred, **kwargs)
-
-    @abstractmethod
-    def _test(self, X, y_pred, **kwargs):
-        """To be implemented by the child class; tests the method on the provided preprocessed dataset.
-
-        Note: _test shouldn"t be directly called by the user, rather it should be called indirectly via the test method.
-
-        Parameters:
-            X (np.ndarray): features, SMILES
-            y_pred (np.ndarray): predicted values
-        """
-        pass
-
-    def test(
+    def build(
         self,
-        dataset: Union[BaseDataset, list, str],
-        use_entire_dataset: bool = False,
-        **kwargs
+        model: BaseModel,
+        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
+        y: Union[np.ndarray, list, pd.Series],
     ):
-        """Evaluates model uncertainty on a test dataset. Calls the _test method on the provided dataset.
+        for error_model in self.error_models:
+            error_model.build(model, X, y)
+        super().build(model, X, y)
 
-        Parameters:
-            dataset (BaseDataset, list, str): dataset being tested, sequence of SMILES
-            use_entire_dataset (bool): whether to use the entire dataset or only the testing data
+    def fit(self, X: Union[pd.DataFrame, np.ndarray, list, pd.Series], y: Union[np.ndarray, list, pd.Series], **kwargs):
+        """Fits confidence scores to an external dataset
+
+        Args:
+            X (array-like): features, smiles
+            y (array-like): true values
         """
-        if isinstance(dataset, BaseDataset):
-            if use_entire_dataset:
-                X = dataset.entire_dataset[0]
-            else:
-                X = dataset.test_dataset[0]
-        elif isinstance(dataset, list):
-            X = dataset
-        else:
-            X = [dataset]
+        y_pred = np.array(self.model.predict(X)).flatten()
+        scores = [error_model.calculate(X, y_pred) for error_model in self.error_models]
+        scores = np.transpose(np.stack(scores))
+        self.reduction.fit(scores)
 
-        y_pred = np.array(self.model.predict(X))
+        residuals = np.abs(np.array(y) - y_pred)
+        scores = self.reduction.transform(scores)
 
-        return self._test(X, y_pred, **kwargs)
+        self._fit(residuals, scores, **kwargs)
 
-    def preprocess(self, X, y=None):
-        """Preprocesses data into the appropriate representation.
+    def calculate(
+        self, X: Union[pd.DataFrame, np.ndarray, list, pd.Series], y_pred: np.ndarray
+    ) -> np.ndarray:
+        """Computes aggregate error model score from inputs.
 
-        The preprocess for the model must return a np.ndarray, e.g. the model must use a
-        BaseStructVecRepresentation or a representation must be passed
-
-        Parameters:
-            X (np.ndarray): features, SMILES
-            y (np.ndarray): values
-
-        Returns:
-            X (np.ndarray): features, representation
+        Args:
+            X: features, smiles
+            y_pred: predicted values
         """
-        if self.rep is None:
-            X = self.model.preprocess(X, y)
-        else:
-            X = np.array(self.rep.convert(X))
+        scores = [error_model.calculate(X, y_pred) for error_model in self.error_models]
+        scores = np.transpose(np.stack(scores))
 
-        assert isinstance(
-            X, np.ndarray
-        ), "The preprocess for the model must return a np.ndarray, e.g. the model must use a BaseCompoundVecRepresentation or a representation must be passed"
-
-        return X
-
-    def _save(self) -> dict:
-        return super()._save()
-
-    def _load(self, d: dict):
-        return super()._load(d)
+        return self.reduction.transform(scores)
